import os
import sys
from multiprocessing import Pool
import atexit
import uuid

from collections import namedtuple

import grass.script as gs
from grass.exceptions import CalledModuleError
from grass.pygrass.modules.grid.split import split_region_tiles


def get_tempory_region(do_cleanup=True):
    """Copies the current region to a temporary region with "g.region save=",
    then returns name of that region. Installs an atexit
    handler to delete the temporary region upon termination by default.
    """
    name = "tmp.%s.%s" % (os.path.basename(sys.argv[0]),
                          str(uuid.uuid4()).replace('-', ''))
    gs.run_command('g.region', save=name, overwrite=True)
    if do_cleanup:
        atexit.register(delete_temporary_region, name)
    return name


def delete_temporary_region(name):
    """Removes any region named by it."""
    try:
        gs.run_command('g.remove', type='region', name=name,
                       flags='f', quiet=True)
    except CalledModuleError:
        pass


def region_to_tiles(width, height, overlap=0, region=None):
    lists = split_region_tiles(region=region, width=width,
                               height=height, overlap=overlap)
    return [to_simple_bbox(item) for sublist in lists for item in sublist]


SimpleBBox = namedtuple('SimpleBBox', 'west, east, north, south')


def to_simple_bbox(bbox):
    return SimpleBBox(west=bbox.west, east=bbox.east,
                      north=bbox.north, south=bbox.south)


def dict_parse_command(*args, **kwargs):
    return dict(gs.parse_command(*args, **kwargs))


# this is basically a list of functors which has a special way of adding
# more items to it
class ModuleCallList(list):

    def __init__(self, bbox=None):
        super(ModuleCallList, self).__init__()
        self.env = None
        if bbox:
            self.bbox = bbox
            self.env = os.environ.copy()
            self.env['WIND_OVERRIDE'] = get_tempory_region()
            bbox = self.bbox
            gs.run_command('g.region', w=bbox.west, e=bbox.east,
                           n=bbox.north, s=bbox.south, env=self.env)

    def run_command(self, *args, **kwargs):
        # TODO: what should happen when env is provided?
        if 'env' not in kwargs:
            kwargs['env'] = self.env
        self.append(ModuleCall(gs.run_command, *args, **kwargs))

    def read_command(self, *args, **kwargs):
        if 'env' not in kwargs:
            kwargs['env'] = self.env
        self.append(ModuleCall(gs.read_command, *args, **kwargs))

    def parse_command(self, *args, **kwargs):
        if 'env' not in kwargs:
            kwargs['env'] = self.env
        # TODO: only normal dict can get pickled, not the GRASS KeyVal
        self.append(ModuleCall(dict_parse_command, *args, **kwargs))

    def mapcalc(self, *args, **kwargs):
        if 'env' not in kwargs:
            kwargs['env'] = self.env
        self.append(ModuleCall(gs.mapcalc, *args, **kwargs))

    def mapcalc3d(self, *args, **kwargs):
        if 'env' not in kwargs:
            kwargs['env'] = self.env
        self.append(ModuleCall(gs.mapcalc3d, *args, **kwargs))

    def execute(self):
        """Execute all queued processes.

        Return whatever was the last return value
        """
        ret = None
        for call in self:
            ret = call.execute()
        return ret


# this is basically a functor
class ModuleCall(object):
    def __init__(self, function, *args, **kwargs):
        self.function = function
        self.args = args
        self.kwargs = kwargs

    def execute(self):
        """Execute all associated process.

        Returns the original return value
        """
        return self.function(*(self.args), **(self.kwargs))


# TODO: do we need the functions twice?
def execute_module_call_list(module_list):
    try:
        return module_list.execute()
    except (KeyboardInterrupt, CalledModuleError):
        return


def execute_module_call(call):
    try:
        return call.execute()
    except (KeyboardInterrupt, CalledModuleError):
        return


# TODO: introduce better naming for modules and functions
# TODO: introduce debug, dry_run and no_clean_tmp wherever appropriate
def execute_list(module_lists, nprocs, debug=False):
    if debug:
        result = []
        for module_list in module_lists:
            result.append(module_list.execute())
        return result
    pool = Pool(nprocs)
    proc = pool.map_async(execute_module_call_list, module_lists)
    # TODO: error handling should be improved
    try:
        return proc.get()
    except (KeyboardInterrupt, CalledModuleError):
        # TODO: this should be done in a better way
        sys.exit("KeyboardInterrupt, CalledModuleError")

# TODO: errors lead to strange behavior, is legal to wrap map_async like this?
def execute_function_on_list(module_lists, function, nprocs, debug=False):
    if debug:
        result = []
        for module_list in module_lists:
            result.append(function(module_list))
        return result
    pool = Pool(nprocs)
    proc = pool.map_async(function, module_lists)
    # TODO: error handling should be improved
    try:
        return proc.get()
    except (KeyboardInterrupt, CalledModuleError):
        # TODO: this should be done in a better way
        sys.exit("KeyboardInterrupt, CalledModuleError")


def execute_by_module(modules, nprocs):
    pool = Pool(nprocs)
    proc = pool.map_async(execute_module_call, modules)
    try:
        return proc.get()
    except (KeyboardInterrupt, CalledModuleError):
        # TODO: this should be done in a better way
        sys.exit("KeyboardInterrupt, CalledModuleError")


def generate_patch_expression(rasters, bboxes):
    exprs = []
    expr = ("if(x() >= {b.west} && x() <= {b.east} "
            "&& y() >= {b.south} && y() <= {b.north}, {m},")
    for raster, bbox in zip(rasters, bboxes):
        exprs.append(expr.format(b=bbox, m=raster))
    return " ".join(exprs) + " null()" + ")" * len(exprs)

# TODO: patch looses color table

def patch_rasters(rasters, bboxes, output):
    expression = generate_patch_expression(rasters, bboxes)
    gs.mapcalc("%s = %s" % (output, expression))


def patch_rasters3d(rasters, bboxes, output):
    expression = generate_patch_expression(rasters, bboxes)
    gs.mapcalc3d("%s = %s" % (output, expression))


def patch_rasters_2(args):
    patch_rasters(*args)


def patch_rasters3d_2(args):
    patch_rasters3d(*args)


def remove_rasters(maps):
    gs.run_command('g.remove', type='raster', name=maps,
                   flags='f', quiet=True)


def remove_rasters3d(maps):
    gs.run_command('g.remove', type='raster3d', name=maps,
                   flags='f', quiet=True)


class Namer(object):
    def __init__(self, bbox, number, callback):
        self.bbox = bbox
        self.number = number  # TODO: this is a propertyand needs a better name
        self.callback = callback

    def name(self, type, name, patch=True):
        tile_name = "%s_%d" % (name, self.number)
        self.callback(type, name, tile_name, self.bbox)
        return tile_name


# this idea for the class is that it is using the general functions
class TiledWorkflow(object):

    def __init__(self, nprocs, width, height, overlap=0, region=None, debug=False):
        self.nprocs = nprocs
        self.bboxes = region_to_tiles(width=width, height=height, overlap=overlap)
        self.bboxes_no_overlap = region_to_tiles(width=width, height=height)
        self.lists_of_rasters = []  # tuple(list, list, str)
        self.lists_of_rasters3d = []  # tuple(list, list, str)
        self.elements_to_patch = {}
        self.calls = []
        self.debug = debug
        self.iteration_index = 0
        self.iteration_max = len(self.bboxes)

    def __iter__(self):
        return self

    def next(self):
        if self.iteration_index >= self.iteration_max:
            raise StopIteration
        else:
            bbox = self.bboxes[self.iteration_index]
            bbox_no_overlap = self.bboxes[self.iteration_index]
            namer = Namer(bbox=bbox_no_overlap, number=self.iteration_index,
                          callback=self.add_element_to_patch)
            call = ModuleCallList(bbox=bbox)
            self.calls.append(call)
            self.iteration_index += 1
            return namer, call

    def execute_tiled(self):
        return execute_list(self.calls, nprocs=self.nprocs, debug=self.debug)

    def add_rasters_to_patch(self, inputs, output):
        if len(self.bboxes) != len(inputs):
            raise ValueError("Number of items to patch must be the same"
                             " as the number of bounding boxes")
        self.lists_of_rasters.append((inputs, self.bboxes_no_overlap, output))

    def add_rasters3d_to_patch(self, inputs, output):
        if len(self.bboxes) != len(inputs):
            raise ValueError("Number of items to patch must be the same"
                             " as the number of bounding boxes")
        self.lists_of_rasters3d.append((inputs, self.bboxes_no_overlap, output))

    def add_element_to_patch(self, type, name, tile_name, bbox):
        if not type in self.elements_to_patch:
            self.elements_to_patch[type] = {}
        if not name in self.elements_to_patch[type]:
            self.elements_to_patch[type][name] = {'tiles': [], 'bboxes': []}
        self.elements_to_patch[type][name]['tiles'].append(tile_name)
        self.elements_to_patch[type][name]['bboxes'].append(bbox)

    def patch(self):
        for type, names in self.elements_to_patch.iteritems():
            if type == 'raster':
                for name, info in names.iteritems():
                    self.lists_of_rasters.append((info['tiles'], info['bboxes'], name))
        
        execute_function_on_list(self.lists_of_rasters, function=patch_rasters_2, nprocs=self.nprocs, debug=self.debug)
        execute_function_on_list(self.lists_of_rasters3d, function=patch_rasters3d_2, nprocs=self.nprocs, debug=self.debug)
        # TODO: parallelize
        # TODO: perhaps special option for temp files, or even preserve tiles
        if not self.debug:
            for rasters, bbox, output in self.lists_of_rasters:
                remove_rasters(rasters)
            for rasters, bbox, output in self.lists_of_rasters3d:
                remove_rasters3d(rasters)

    def execute(self):
        result = self.execute_tiled()
        self.patch()
        return result


def main():
    nprocs = 4
    width = 500
    height = 500
    overlap = 10

    tiled_workflow = TiledWorkflow(nprocs=nprocs, width=width, height=height, overlap=overlap)

    print "Splitting processing into %s tiles" % len(tiled_workflow.bboxes)

    aspects = []
    slopes = []
    for namer, workflow in tiled_workflow:
        i = namer.number
        slope = 'slope_' + str(i)
        aspect = 'aspect_' + str(i)

        workflow.run_command('r.slope.aspect', elevation='fractal_surface',
                             slope=slope)
        workflow.run_command('r.slope.aspect', elevation='fractal_surface',
                             aspect=aspect)
        workflow.parse_command('g.region', flags='pg')

        slopes.append(slope)
        aspects.append(aspect)

    results = tiled_workflow.execute_tiled()
    tiled_workflow.add_rasters_to_patch(slopes, 'slope_patched')
    tiled_workflow.add_rasters_to_patch(aspects, 'aspect_patched')
    tiled_workflow.patch()

    for result in results:
        print "-" * 72
        for key, value in result.iteritems():
            print key, value

    nprocs = 4
    width = 500
    height = 500
    overlap = 10

    tiled_workflow = TiledWorkflow(nprocs=nprocs, width=width, height=height, overlap=overlap)

    print "Splitting processing into %s tiles" % len(tiled_workflow.bboxes)

    for namer, workflow in tiled_workflow:
        slope = namer.name('raster', 'slope2')
        aspect = namer.name('raster','aspect2')

        workflow.run_command('r.slope.aspect', elevation='fractal_surface',
                             slope=slope)
        workflow.run_command('r.slope.aspect', elevation='fractal_surface',
                             aspect=aspect)
        workflow.parse_command('g.region', flags='pg')

    results = tiled_workflow.execute()

    for result in results:
        print "-" * 72
        for key, value in result.iteritems():
            print key, value


if __name__ == "__main__":
    sys.exit(main())
